@@ -25,7 +25,7 @@ def get_versioner(single_object_qs, versions_type, **kwargs):
     versions = Version.objects.get_for_object(instance)
 
     return versions_type(
-        objects=versions,
+        objects=list(versions), #R.map(lambda version: version._object_version.object, list(versions)),
         **kwargs
     )
 
@@ -59,13 +59,20 @@ class VersionType(DjangoObjectType):
 def create_version_type(model_object_type, model_object_type_fields):
     # We can't assign Version as the Meta model because multiple classes would point at the same model,
     # which probably isn't allowed
-    versioned_type_model = type(
-        f'VersionedTypeMixinFor{model_object_type.__name__}',
+
+    def resolve_instance(parent, info, **kwargs):
+        instance = parent._object_version.object
+        # Inject the version so RevisionModelMixin knows how to handle
+        instance._version = parent
+        return instance
+
+    version_type_model = type(
+        f'VersionTypeModelFor{model_object_type.__name__}',
         (ObjectType,),
         dict(
             id=Int(),
-            reversion=Field(RevisionType),
-            field_dict=Field(model_object_type_fields)
+            revision=Field(RevisionType),
+            instance=Field(model_object_type, resolver=resolve_instance)
         )
     )
 
@@ -78,12 +85,17 @@ def create_version_type(model_object_type, model_object_type_fields):
             fields=revision_fields,
             type_modifier=lambda *type_and_args: Field(*type_and_args)
         ),
-        field_dict=model_object_type_fields
+        instance=dict(
+            type=model_object_type,
+            graphene_type=model_object_type,
+            fields=model_object_type_fields,
+            type_modifier=lambda *type_and_args: Field(*type_and_args)
+        )
     ))
-    return dict(type=versioned_type_model, fields=versioned_fields)
+    return dict(type=version_type_model, fields=versioned_fields)
 
 
-def create_versions_type(model_object_type, model_object_type_fields):
+def create_version_container_type(model_object_type, model_object_type_fields):
     """
         DjangObjectType and fields to hold all the versions of one instance
     :param model_object_type:
@@ -96,10 +108,10 @@ def create_versions_type(model_object_type, model_object_type_fields):
     )
 
     versions_type_model = type(
-        f'VersionsTypeModelFor{model_object_type.__name__}',
+        f'VersionContainerTypeModelFor{model_object_type.__name__}',
         (ObjectType,),
         dict(
-            objects=List(version_type),
+            objects=List(version_type)
         )
     )
 
