@@ -4,25 +4,24 @@ import logging
 import sys
 from decimal import Decimal
 
+import graphene
 import reversion
 from deepmerge import Merger
-from graphql.language.printer import print_ast
-import graphene
 from django.contrib.gis.db.models import GeometryField, OneToOneField, ManyToManyField, ForeignKey, \
     GeometryCollectionField
-from graphql import parse
-
-from rescape_python_helpers import ramda as R, memoize
 from django.contrib.postgres.fields import JSONField
 from django.db.models import AutoField, CharField, BooleanField, BigAutoField, DecimalField, \
     DateTimeField, DateField, BinaryField, TimeField, FloatField, EmailField, UUIDField, TextField, IntegerField, \
     BigIntegerField, NullBooleanField, Q
 from graphene import Scalar, InputObjectType, ObjectType
+from graphql import parse
 from graphql.language import ast
+from graphql.language.printer import print_ast
 from inflection import camelize
+from rescape_python_helpers import ramda as R, memoize
 from rescape_python_helpers.functional.ramda import to_dict_deep, flatten_dct, to_pairs, flatten_dct_until
 
-from .graphene_helpers import dump_graphql_keys, dump_graphql_data_object
+from .graphene_helpers import dump_graphql_keys, dump_graphql_data_object, camelize_graphql_data_object, call_if_lambda
 
 logger = logging.getLogger('rescape_graphene')
 from django.conf import settings
@@ -157,16 +156,19 @@ class DataPointRelatedCreateInputType(InputObjectType):
     id = graphene.String(required=True)
 
 
-@memoize(map_args=lambda args: [
+def _memoize(args):
+   return [
     # Only use graphene_type here. type is a function and can't be serialized
     args[0]['graphene_type'],
     args[1],
     # TODO We use the parent_type_class to make each type unique. I don't know why graphene won't let us reuse
     # input types within the schema. It seems like a UserInputType should be reusable whether it's the User
     # of a Region or the user of a Group.
-    args[2] if R.isinstance((list, tuple), args[2]) else [args[2]]
-])
-def input_type_class(field_config, crud, parent_type_classes=[]):
+    args[2] if R.lte(3, R.length(args)) and R.isinstance((list, tuple), args[2]) else [args[2]],
+    args[3] if R.lte(2, R.length(args)) else False
+]
+@memoize(map_args=_memoize)
+def input_type_class(field_config, crud, parent_type_classes=[], allowed_fields_only=False):
     """
     An InputObjectType subclass for use as nested query argument types and mutation argument types
     The subclass is dynamically created based on the field_dict_value['graphene_type'] and the crud type.
@@ -182,7 +184,9 @@ def input_type_class(field_config, crud, parent_type_classes=[]):
     # Get the Graphene type. This comes from graphene_type if the class containing the field is a Django Model,
     # It defaults to type, which is what we expect if we didn't have to use a graphene_type to distinguish
     # from the underlying Django type
-    graphene_class = field_config['graphene_type'] or field_config['type']
+    graphene_class = call_if_lambda(field_config['graphene_type'] or field_config['type'])
+    fields = call_if_lambda(field_config['fields'])
+
     # Make it an array if not
     modified_parent_type_classes = parent_type_classes if R.isinstance((list, tuple), parent_type_classes) else [
         parent_type_classes]
@@ -198,10 +202,16 @@ def input_type_class(field_config, crud, parent_type_classes=[]):
             # such as referenced_in_create = REQUIRE, if the container has to have a reference when the
             # container is created
             R.map_with_obj(lambda key, value: R.omit(['create', 'update'], value)),
-            # Only take the id
-            R.pick(['id'])
-        )(field_config['fields']) if crud in [CREATE, UPDATE] else field_config['fields']
-    ) if django_model_of_graphene_type(graphene_class) else field_config['fields']
+            # Only take the id, unless related_input=true for a field
+            lambda fields: R.merge(
+                R.pick(['id'], fields),
+                R.filter_dict(
+                    lambda name_field: R.compose(R.equals(ALLOW), R.prop_or(False, 'related_input'))(name_field[1]),
+                    fields
+                )
+            )
+        )(fields) if crud in [CREATE, UPDATE] else fields
+    ) if django_model_of_graphene_type(graphene_class) else fields
 
     input_fields = input_type_fields(
         input_type_field_configs,
@@ -214,6 +224,10 @@ def input_type_class(field_config, crud, parent_type_classes=[]):
     # This doesn't apply to Update and Create input types, since we never filter during those operations
     filter_fields = allowed_filter_arguments(input_type_field_configs, graphene_class) if R.equals(READ, crud) else {}
 
+    combined_fields = R.merge(filter_fields, input_fields)
+    if allowed_fields_only:
+        return combined_fields
+
     return type(
         '%s%sRelated%sInputType' % (
             graphene_class.__name__,
@@ -230,7 +244,7 @@ def input_type_class(field_config, crud, parent_type_classes=[]):
         # Django object.
         #
         # Otherwise field_dict_value['fields'] are independent of a Django model and each have their own type property
-        R.merge(input_fields, filter_fields)
+        combined_fields
     )
 
 
@@ -330,9 +344,9 @@ def process_field(field_to_unique_field_groups, field, field_dict_value, parent_
     :return: A dict with the unique property and anything else we need
     """
     unique = R.compact([
-        PRIMARY if field.primary_key else None,
-        UNIQUE if field.unique else None,
-        R.prop_or(None, field.attname, field_to_unique_field_groups)
+        PRIMARY if R.has('primary_key', field) else None,
+        UNIQUE if R.has('unique', field) else None,
+        R.prop_or(None, R.has('attname', field), field_to_unique_field_groups)
     ])
     # Normally the field_dict_value will delegate the type to the underlying Django model
     # In cases where we need an explicit type, because the field represents something modeled outside django,
@@ -379,7 +393,7 @@ def parse_django_class(model, field_dict, parent_type_classes=[]):
         # Only accept model fields that are defined in field_dict
         R.filter(
             lambda field: field.name in field_dict,
-            R.concat(model._meta.fields, model._meta.many_to_many)
+            R.concat(model._meta.fields, R.concat(model._meta.many_to_many, model._meta.related_objects))
         )
     ))
 
@@ -497,6 +511,18 @@ def add_filters(argument_dict):
     )(argument_dict)
 
 
+def top_level_allowed_filter_arguments(fields, graphene_type):
+    """
+        For top-level calls, not recursion
+        Like allowed_query_and_read_arguments but only used for arguments and adds filter variables like id_contains.
+        Note that django needs __ so these are converted for resolvers. The graphql interface converts them to
+        camel case
+    :param fields: The fields for the graphene type
+    :param graphen_type: The graphene type
+    :return: dict of field keys and there graphene type, either a primitive or input type
+    """
+    return input_type_class(dict(fields=fields, graphene_type=graphene_type), 'read', [], True)
+
 def allowed_filter_arguments(fields_dict, graphene_type):
     """
         Like allowed_query_and_read_arguments but only used for arguments and adds filter variables like id_contains.
@@ -510,14 +536,13 @@ def allowed_filter_arguments(fields_dict, graphene_type):
         lambda argument_dict: add_filters(argument_dict),
         R.map_dict(resolve_type(graphene_type)),
         R.filter_dict(
-            lambda key_value:
             # Don't allow DENYed READ fields to be used for querying
-            R.and_func(
+            lambda key_value: R.and_func(
                 True,
                 R.not_func(R.prop_eq_or_in(READ, DENY, key_value[1]))
             )
         )
-    )(fields_dict)
+    )(call_if_lambda(fields_dict))
 
 
 def process_filter_kwargs(model, kwargs):
@@ -577,8 +602,18 @@ def instantiate_graphene_type(field_config, parent_type_classes, crud):
         fields = R.prop('fields', field_config)
         resolved_graphene_type = input_type_class(dict(graphene_type=graphene_type, fields=fields), crud,
                                                   parent_type_classes)
+    elif R.isfunction(graphene_type) and \
+            R.compose(R.equals(0), R.length, R.prop('parameters'), inspect.signature)(graphene_type):
+        # If out graphene_type is a no arg lambda it means it needs lazy evaluation to avoid circular imports
+        # This is only true for representing reverse relationships on django models
+        _graphene_type = graphene_type()
+        _fields = R.compose(call_if_lambda, R.prop('fields'))(field_config)
+        resolved_graphene_type = input_type_class(
+            dict(graphene_type=_graphene_type, fields=_fields), crud,
+            parent_type_classes
+        )
     elif R.isfunction(graphene_type):
-        # If a lambda is returned we have an InputType subclass that needs to know the crud type
+        # If a lambda is returned with params we have an InputType subclass that needs to know the crud type
         resolved_graphene_type = graphene_type(crud)
     else:
         # Otherwise we have a simple type
@@ -595,6 +630,7 @@ def instantiate_graphene_type(field_config, parent_type_classes, crud):
         )
 
 
+
 def input_type_fields(fields_dict, crud, parent_type_classes=[]):
     """
     :param fields_dict: The fields_dict for the Django model or json data
@@ -609,7 +645,7 @@ def input_type_fields(fields_dict, crud, parent_type_classes=[]):
         # This means that if the user tries to pass these fields to graphql an error will occur
         R.filter_dict(
             lambda key_value: R.not_func(R.prop_eq_or(False, crud, DENY, key_value[1])),
-            fields_dict
+            call_if_lambda(fields_dict)
         )
     )
 
@@ -715,7 +751,7 @@ def graphql_query(graphene_type, fields, query_name):
     The only key allowed is variables, which contains param key values. Example: variables={'user': 'Peter'}
     This results in query whatever(id: String!) { query_name(id: id) ... }
     """
-    field_type_lookup = allowed_filter_arguments(fields, graphene_type)
+    field_type_lookup = top_level_allowed_filter_arguments(fields, graphene_type)
 
     def form_query(client, field_overrides={}, **kwargs):
         """
@@ -728,7 +764,16 @@ def graphql_query(graphene_type, fields, query_name):
         # Map the field_type_lookup to the right graphene type, either a primitive like 'String' or a complex
         # read input type like 'FeatureCollectionDataTypeofFooTypeRelatedReadInputType'
         variable_definitions = R.map_key_values(
-            lambda k, v: [camelize(k, False), field_type_lookup[k]._meta.name],
+            lambda k, v: [
+                camelize(k, False),
+                R.if_else(
+                    lambda lookup: R.has('_meta', lookup),
+                    # Field Case
+                    lambda lookup: lookup._meta.name,
+                    # List Case
+                    lambda lookup: f'[{lookup._of_type._meta.name}]'
+                )(field_type_lookup[k])
+            ],
             kwargs['variables']
         ) if R.has('variables', kwargs) else {}
 
@@ -757,7 +802,7 @@ def graphql_query(graphene_type, fields, query_name):
                     lambda key: '%s: $%s' % (key, key),
                     R.keys(variable_definitions))
             ) if variable_definitions else '',
-            dump_graphql_keys(field_overrides or fields)
+            dump_graphql_keys(field_overrides or call_if_lambda(fields))
         )))
 
         # Update the variable names to have camel case instead of pythonic slugs
@@ -796,6 +841,15 @@ def grapqhl_authorization_mutation(client, values):
     client.execute(mutation, variables=values)
 
 
+def capitalize_first_letter(str):
+    """
+    Capitalize the first letter since str.capitalize() lowercases everything first (yes Python sucks)
+    :param str:
+    :return:
+    """
+    return str[:1].upper() + str[1:]
+
+
 @R.curry
 def graphql_update_or_create(mutation_config, fields, client, values):
     """
@@ -823,8 +877,8 @@ def graphql_update_or_create(mutation_config, fields, client, values):
     # where className is the camel-case version of the given class name in mutation_config.class_name
     name = camelize(R.prop('class_name', mutation_config), False)
     mutation = print_ast(parse(''' 
-        mutation %sMutation {
-            %s(%sData: %s) {
+        mutation %sMutation($data: %sInputType!) {
+            %s(%sData: $data) {
                 %s {
                     %s 
                 }
@@ -836,12 +890,10 @@ def graphql_update_or_create(mutation_config, fields, client, values):
         # Actual schema function which matches something in the schema
         # This will be createClass or updateClass where class is the class name e.g. createFoo or updateFoo
         # Keep in mind that in python in the schema it will be defined create_foo or update_foo
+        capitalize_first_letter(R.item_path(['crud', update_or_create], mutation_config)),
         R.item_path(['crud', update_or_create], mutation_config),
         # The name of the InputDataType that is defined for this function, e.g. FooInputDataType
         name,
-        # Key values for what is being created or updated. This is dumped recursively and matches the structure
-        # of the InputDataType subclass
-        dump_graphql_data_object(values),
         # Again the name, this time used for the structure of the return query e.g. foo { ...return value ...}
         name,
         # The return query dump, which are all the fields available that aren't marked read=IGNORE.
@@ -849,8 +901,11 @@ def graphql_update_or_create(mutation_config, fields, client, values):
         # which isn't part of field_configs because Graphene handles ids automatically
         dump_graphql_keys(R.merge(dict(id=dict(type=graphene.Int)), fields)),
     )))
-    logger.debug('Mutation: %s' % mutation)
-    return client.execute(mutation)
+    # Key values for what is being created or updated. This is dumped recursively and matches the structure
+    # of the InputDataType subclass
+    variables = dump_graphql_data_object(dict(data=values))
+    logger.debug(f'Mutation: {mutation}\nVariables: {variables}')
+    return client.execute(mutation, variables=camelize_graphql_data_object(dict(data=values)))
 
 
 def process_query_value(model, value_dict):
@@ -870,13 +925,15 @@ def process_query_value(model, value_dict):
         )
     )
 
+def not_ends_with_contains(key):
+    return not key.endswith('contains')
 
 def process_query_kwarg(model, key, value):
     """
         Process a query kwarg. The key is always a string and the value can be a scalar or a dict representing the
-        model given by model. E.g. model: User, key: 'user', value: {id: 1, username: 'jo'} or
+        given model. E.g. model: User, key: 'user', value: {id: 1, username: 'jo'} or
         model User, key: 'user', value: {id: 1, group: {id: 2}}
-        TODO I haven't made this work with ManyToMany yet, such as
+        This works in a limited capacity with ManyToMany. You can pass id__in=[{id: x1, id: x2, ...}]
         model User, key: 'user', value: {id: 1, groups: [{id: 2}, {name: 'fellas'}]}. This requires using __in
         for the django key and other fancy stuff
     :param model:
@@ -900,14 +957,15 @@ def process_query_kwarg(model, key, value):
         # This is just one way of filtering json. We can also do it with the argument structure
         return R.compose(
             lambda dct: R.map_with_obj_to_values(lambda key, value: Q(**{key: value}), dct),
-            lambda dct: flatten_dct_until(dct, lambda key: not key.endswith('contains'), '__')
+            lambda dct: flatten_dct_until(dct, lambda key: not isinstance(str, key) and not_ends_with_contains(key), '__')
         )({key: value})
     elif R.has(key, model._meta._forward_fields_map):
         # If it's a model key
-        if isinstance(model._meta._forward_fields_map[key], (ForeignKey, OneToOneField)):
+        if isinstance(model._meta._forward_fields_map[key], (ForeignKey, OneToOneField, ManyToManyField)):
             # Recurse on these, so foo: {bar: 1, car: 2} resolves to [['foo__bar' 1], ['foo__car', 2]]
             related_model = model._meta._forward_fields_map[key].related_model
-            q_expressions = process_query_value(related_model, value)
+            # values have to be arrays, so iterate over them
+            q_expressions = R.chain(lambda v: process_query_value(related_model, v), value)
             return R.map(
                 # TODO we assume simple Q expressions with ony one child at children[0]
                 lambda q_expression: Q(**{f'{key}__{q_expression.children[0][0]}': q_expression.children[0][1]}),
@@ -915,6 +973,11 @@ def process_query_kwarg(model, key, value):
             )
         elif isinstance(model._meta._forward_fields_map[key], (ManyToManyField)):
             raise NotImplementedError("TODO need to implement stringify for ManyToMany")
+    elif R.contains(R.head(key.split('__')), R.map(R.prop('name'), model._meta.related_objects)):
+        # Reverse Many-to-Many relationships, the only thing I know how to support at this point is
+        # __in=[ids], although it might be possible to support other filters
+        # If ids aren't in the value list objects, this will fail
+        return [Q(**{key: R.map(R.prop('id'), value)})]
 
     return [Q(**{key: value})]
 
